#ifndef Py_COMPILE_H
#define Py_COMPILE_H

#ifdef __cplusplus
extern "C" {
#endif

/* These constants are also defined in cpyext/eval.py */
#define Py_single_input 256
#define Py_file_input 257
#define Py_eval_input 258
#define Py_func_type_input 345

#ifdef __cplusplus
}
#endif
#endif /* !Py_COMPILE_H */
